@@ -13,6 +13,9 @@ except ImportError as e:
 
 from Onboard.utils        import Rect, EventSource, Process, unicode_str
 
+# Config Singleton
+from Onboard.Config import Config
+config = Config()
 
 class AsyncEvent:
     """
@@ -91,9 +94,15 @@ class AtspiStateTracker(EventSource):
         self._update_listeners()
 
     def disconnect(self, event_name, callback):
+        had_listeners = self.has_listeners(self._event_names)
+
         EventSource.disconnect(self, event_name, callback)
         self._update_listeners()
 
+        # help debugging disconnecting events on exit
+        if had_listeners and not self.has_listeners(self._event_names):
+            _logger.info("all listeners disconnected")
+
     def _update_listeners(self):
         register = self.has_listeners(self._focus_event_names)
         self._register_atspi_focus_listeners(register)
@@ -269,7 +278,7 @@ class AtspiStateTracker(EventSource):
 
     def _on_atspi_keystroke(self, event, user_data):
         if event.type == Atspi.EventType.KEY_PRESSED_EVENT:
-            _logger.debug("key-stroke {} {} {} {}" \
+            _logger.atspi("key-stroke {} {} {} {}" \
                           .format(event.modifiers,
                                   event.hw_code, event.id, event.is_text))
             #keysym = event.id # What is this? Not XK_ keysyms apparently.
@@ -297,7 +306,7 @@ class AtspiStateTracker(EventSource):
                             self._read_initial_accessible_state(accessible)
                 except Exception as ex: # Private exception gi._glib.GError when
                                         # gedit became unresponsive.
-                    _logger.info("_on_focus_changed(): "
+                    _logger.atspi("_on_focus_changed(): "
                                  "invalid accessible, failed to read state: " \
                                  + unicode_str(ex))
 
@@ -331,7 +340,7 @@ class AtspiStateTracker(EventSource):
                                 self._read_remaining_accessible_state(accessible))
                 except Exception as ex: # Private exception gi._glib.GError when
                                         # gedit became unresponsive.
-                    _logger.info("_set_active_accessible(): "
+                    _logger.atspi("_set_active_accessible(): "
                             "invalid accessible, failed to read remaining state: " \
                             + unicode_str(ex))
 
@@ -380,18 +389,38 @@ class AtspiStateTracker(EventSource):
         """ Screen rect of the active accessible """
         return self._state.get("extents", Rect())
 
+    @staticmethod
+    def _get_accessible_extents(accessible):
+        """
+        Screen rect of the given accessible, no caching,
+        no exception handling.
+        """
+        scale = config.window_scaling_factor
+        if scale != 1.0:
+            attributes = accessible.get_attributes()
+            # Only Gtk-3 widgets return scaled coordinates, all others,
+            # including Gtk-2 apps like firefox, clawsmail and Qt-apps,
+            # apparently don't.
+            if AtspiStateTracker.is_toolkit_gtk3(attributes):
+                scale = 1.0
+            else:
+                scale = 1.0 / config.window_scaling_factor
+
+        ext = accessible.get_extents(Atspi.CoordType.SCREEN)
+        return Rect(ext.x * scale, ext.y * scale,
+                    ext.width * scale, ext.height * scale)
+
     @staticmethod
     def get_accessible_extents(accessible):
         """ Screen rect of the given accessible, no caching """
         try:
-            ext = accessible.get_extents(Atspi.CoordType.SCREEN)
+            rect = AtspiStateTracker._get_accessible_extents(accessible)
         except Exception as ex: # private exception gi._glib.GError when
                 # right clicking onboards unity2d launcher (Precise)
-            _logger.info("Invalid accessible,"
+            _logger.atspi("Invalid accessible,"
                          " failed to get extents: " + unicode_str(ex))
-            return Rect()
-
-        return Rect(ext.x, ext.y, ext.width, ext.height)
+            rect = Rect()
+        return rect
 
     @staticmethod
     def get_accessible_text(accessible, begin, end):
@@ -401,7 +430,7 @@ class AtspiStateTracker(EventSource):
         except Exception as ex: # private exception
                                 # gi._glib.GError: timeout from dbind
                                 # with web search in firefox.
-            _logger.info("Invalid accessible,"
+            _logger.atspi("Invalid accessible,"
                          " failed to get text: " + unicode_str(ex))
             return None
 
@@ -453,12 +482,10 @@ class AtspiStateTracker(EventSource):
 
         state["attributes"] = accessible.get_attributes()
         state["interfaces"] = accessible.get_interfaces()
-
-        ext = accessible.get_extents(Atspi.CoordType.SCREEN)
-        state["extents"] = Rect(ext.x, ext.y, ext.width, ext.height)
+        state["extents"] = self._get_accessible_extents(accessible)
 
         # These are currently used only in debug output
-        if _logger.isEnabledFor(logging.DEBUG):
+        if _logger.isEnabledFor(_logger.LEVEL_ATSPI):
             state["id"] = accessible.get_id()
             state["name"] = accessible.get_name()
             pid = accessible.get_process_id()
@@ -473,8 +500,14 @@ class AtspiStateTracker(EventSource):
 
         return state
 
+    @staticmethod
+    def is_toolkit_gtk3(attributes):
+        """ Are the accessible attributes from a gtk3 widget? """
+        return attributes and \
+               "toolkit" in attributes and attributes["toolkit"] == "gtk"
+
     def _log_accessible(self, accessible, focused):
-        if _logger.isEnabledFor(logging.DEBUG):
+        if _logger.isEnabledFor(_logger.LEVEL_ATSPI):
             msg = "AT-SPI focus event: focused={}, ".format(focused)
             if not accessible:
                 msg += "accessible={}".format(accessible)
@@ -494,8 +527,7 @@ class AtspiStateTracker(EventSource):
                     states = state_set.states
                     editable = state_set.contains(Atspi.StateType.EDITABLE) \
                                if state_set else None
-                    ext = accessible.get_extents(Atspi.CoordType.SCREEN)
-                    extents   = Rect(ext.x, ext.y, ext.width, ext.height)
+                    extents = self._get_accessible_extents(accessible)
                 except: # private exception gi._glib.GError when gedit became unresponsive
                     pass
 
@@ -503,13 +535,13 @@ class AtspiStateTracker(EventSource):
                        "editable={editable}, states={states}, " \
                        "extents={extents}]" \
                         .format(name=name,
-                                role = role,
+                                role = role.value_name if role else role,
                                 role_name = role_name,
                                 editable = editable,
                                 states = states,
                                 extents = extents \
                                )
-            _logger.debug(msg)
+            _logger.atspi(msg)
 
 
 class AtspiStateType:
