@@ -5,7 +5,8 @@ from __future__ import division, print_function, unicode_literals
 import time
 from gi.repository import GObject, GLib, GdkX11, Gdk, Gtk
 
-from Onboard.utils       import Rect, CallOnce, Timer
+from Onboard.definitions import DockingMonitor
+from Onboard.utils       import Rect, CallOnce, Timer, gtk_has_resize_grip_support
 from Onboard.WindowUtils import Orientation, WindowRectPersist, \
                                 set_unity_property
 import Onboard.osk as osk
@@ -21,6 +22,7 @@ config = Config()
 ########################
 
 
+
 class KbdWindowBase:
     """
     Very messy class holds the keyboard widget. The mess is the docked
@@ -48,13 +50,19 @@ class KbdWindowBase:
         self._docking_enabled = False
         self._docking_edge = None
         self._docking_rect = Rect()
+        self._docking_monitor = None
         self._shrink_work_area = False
         self._dock_expand = False
         self._current_struts = None
+
+        self._current_docking_monitor = None
         self._monitor_workarea = {}
 
         self._opacity = 1.0
-        self._default_resize_grip = self.get_has_resize_grip()
+        if gtk_has_resize_grip_support():
+            self._default_resize_grip = self.get_has_resize_grip()
+        else:
+            self._default_resize_grip = False
         self._type_hint = None
 
         self._known_window_rects = []
@@ -253,10 +261,11 @@ class KbdWindowBase:
                 self._recreate_window()
 
             # Show the resize gripper?
-            if config.has_window_decoration():
-                self.set_has_resize_grip(self._default_resize_grip)
-            else:
-                self.set_has_resize_grip(False)
+            if gtk_has_resize_grip_support():
+                if config.has_window_decoration():
+                    self.set_has_resize_grip(self._default_resize_grip)
+                else:
+                    self.set_has_resize_grip(False)
 
             self.update_sticky_state()
 
@@ -314,6 +323,10 @@ class KbdWindowBase:
 
     def on_visibility_changed(self, visible):
 
+        if not visible:
+            keyboard = self.keyboard_widget.keyboard
+            keyboard.stop_raise_attempts()
+
         visible_before = self._visible
         self._visible = visible
 
@@ -423,7 +436,8 @@ class KbdWindowBase:
                 # - Unminimizing from unity-2d launcher is a user
                 #   triggered unhide -> lock auto-show visible.
                 # - Minimizing while locked visible -> unlock
-                self.keyboard_widget.lock_auto_show_visible(visible)
+                keyboard = self.keyboard_widget.keyboard
+                keyboard.update_auto_show_on_visibility_change(visible)
 
             return
 
@@ -474,7 +488,6 @@ class KbdWindowBase:
             if biggestHeight < tempHeight:
                 biggestHeight = tempHeight
 
-        geom = self.get_screen().get_monitor_geometry(0)
         eg = self.edgeGravity
         x, y = self.window.get_origin()
 
@@ -484,8 +497,6 @@ class KbdWindowBase:
             propvals[2] = height + y
             propvals[9] = width
         elif eg == Gdk.Gravity.SOUTH and y != 0:
-            #propvals[2] = y
-            #propvals[9] = geom.width - 1
             propvals[3] = biggestHeight - y
             propvals[11] = width - 1
 
@@ -567,9 +578,9 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
 
            if config.is_docking_enabled() and \
               not config.xid_mode:
-                mon = self.get_docking_monitor()
-                new_area = self.get_monitor_workarea()
-                area = self._monitor_workarea.get(0)
+                monitor = self.get_docking_monitor()
+                new_area = self.get_monitor_workarea(monitor)
+                area = self._monitor_workarea.get(monitor)
                 if area:
                     # Only check for x changes, y is too dangerous for now,
                     # too easy to get the timing wrong and end up with double docks.
@@ -645,9 +656,12 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
                 self.restore_window_rect()
 
     def on_user_positioning_begin(self):
+        keyboard = self.keyboard_widget.keyboard
+
         self.stop_save_position_timer()
-        self.stop_auto_position()
-        self.keyboard_widget.freeze_auto_show()
+        self.stop_auto_positioning()
+        keyboard.stop_raise_attempts()
+        keyboard.freeze_auto_show()
 
     def on_user_positioning_done(self):
         self.update_window_rect()
@@ -662,7 +676,8 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
 
         # Thaw auto show after a short delay to stop the window
         # from hiding due to spurios focus events after a system resize.
-        self.keyboard_widget.thaw_auto_show(1.0)
+        keyboard = self.keyboard_widget.keyboard
+        keyboard.thaw_auto_show(1.0)
 
     def detect_docking(self):
         if self.keyboard_widget.was_moving():
@@ -896,7 +911,9 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
 
     def auto_position(self):
         self._auto_position_started = True
-        self.update_position()
+
+        # Leave the first commit to AutoShow._begin_transition()
+        self.update_position(False)
 
         # With docking enabled, when focusing the search entry of a
         # maximized firefox window, it changes position when the work
@@ -922,10 +939,10 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         else:
             return False
 
-    def stop_auto_position(self):
+    def stop_auto_positioning(self):
         self._auto_position_poll_timer.stop()
 
-    def update_position(self):
+    def update_position(self, commit=True):
         home_rect = self.get_home_rect()
         rect = self.get_repositioned_window_rect(home_rect)
         if rect is None:
@@ -934,29 +951,37 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
 
         if self.get_position() != rect.get_position():
             self.keyboard_widget.transition_position_to(rect.x, rect.y)
-            self.keyboard_widget.commit_transition()
+            if commit:
+                self.keyboard_widget.commit_transition()
 
     def get_repositioned_window_rect(self, home_rect):
-        clearance = config.auto_show.widget_clearance
-        test_clearance = clearance
-        move_clearance = clearance
-        limit_rects = None  # None: all monitors
+        if config.can_auto_show_reposition():
+            clearance = config.auto_show.widget_clearance
+            test_clearance = clearance
+            move_clearance = clearance
+            limit_rects = None  # None: all monitors
 
-        # No test clearance when docking. Make it harder to jump
-        # out of the dock, for example for the bottom search box
-        # in maximized firefox.
-        if config.is_docking_enabled():
-            test_clearance = (clearance[0], 0, clearance[2], 0)
+            # No test clearance when docking. Make it harder to jump
+            # out of the dock, for example for the bottom search box
+            # in maximized firefox.
+            if config.is_docking_enabled():
+                test_clearance = (clearance[0], 0, clearance[2], 0)
 
-            # limit the horizontal freedom to the docking monitor
-            area, geom = self.get_docking_monitor_rects()
-            limit_rects = [area]
+                # limit the horizontal freedom to the docking monitor
+                area, geom = self.get_docking_monitor_rects()
+                limit_rects = [area]
 
-        horizontal, vertical = self.get_repositioning_constraints()
-        return self.keyboard_widget.auto_show.get_repositioned_window_rect( \
-                                        home_rect, limit_rects,
-                                        test_clearance, move_clearance,
-                                        horizontal, vertical)
+            horizontal, vertical = self.get_repositioning_constraints()
+            keyboard = self.keyboard_widget.keyboard
+            rect = keyboard.get_auto_show_repositioned_window_rect( \
+                                            self.keyboard_widget,
+                                            home_rect, limit_rects,
+                                            test_clearance, move_clearance,
+                                            horizontal, vertical)
+        else:
+            rect = None
+
+        return rect
 
     def reposition(self, x, y):
         """
@@ -1001,7 +1026,7 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
 
     def get_hidden_rect(self):
         """
-        Returns the rect of the hidden window rect with auto-show
+        Returns the rect of the hidden window with auto-show
         repositioning taken into account.
         """
         if config.is_docking_enabled():
@@ -1155,7 +1180,7 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
 
     def on_docking_notify(self):
         self.update_docking()
-        self.keyboard_widget.update_resize_handles()
+        self.keyboard_widget.update_window_handles()
 
     def update_docking(self, force_update = False):
         enable = config.is_docking_enabled()
@@ -1164,25 +1189,30 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         else:
             rect = Rect()
         shrink = config.window.docking_shrink_workarea
-        edge = config.window.docking_edge
         expand = self.get_dock_expand()
+        edge = config.window.docking_edge
+        monitor = self.get_docking_monitor(True)
 
         if self._docking_enabled != enable or \
            (self._docking_enabled and \
             (self._docking_rect != rect or \
              self._shrink_work_area != shrink or \
              self._dock_expand != expand or \
-             bool(self._current_struts) != shrink)
+             bool(self._current_struts) != shrink or
+             self._docking_monitor != monitor)
            ):
-            self.enable_docking(enable)
+            self._current_docking_monitor = monitor
+
+            self._realize_docking(enable)
 
             self._shrink_work_area = shrink
             self._dock_expand = expand
             self._docking_edge = edge
+            self._docking_monitor = monitor
             self._docking_enabled = enable
             self._docking_rect = rect
 
-    def enable_docking(self, enable):
+    def _realize_docking(self, enable):
         if enable:
             self._set_docking_struts(config.window.docking_shrink_workarea,
                                      config.window.docking_edge,
@@ -1192,6 +1222,12 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
             self.restore_window_rect()
             self.clear_struts()
 
+    def update_docking_monitor(self):
+        mon_before = self._current_docking_monitor
+        mon_now = self.get_docking_monitor(True)
+        self._current_docking_monitor = mon_now
+        return mon_before != mon_now
+
     def clear_struts(self):
         self._set_docking_struts(False)
 
@@ -1213,7 +1249,6 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
         rect = self.get_dock_rect()
         top_start_x = top_end_x = 0
         bottom_start_x = bottom_end_x = 0
-        #print("geom", geom, "area", area, "rect", rect)
 
         if edge: # Bottom
             top    = 0
@@ -1294,38 +1329,64 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
 
     def get_docking_monitor_rects(self):
         screen = self.get_screen()
-        mon = self.get_docking_monitor()
+        monitor = self.get_docking_monitor()
 
-        area = self._monitor_workarea.get(mon)
+        area = self._monitor_workarea.get(monitor)
         if area is None:
-            area = self.update_monitor_workarea()
+            area = self.update_monitor_workarea(monitor)
 
-        geom = screen.get_monitor_geometry(mon)
+        geom = screen.get_monitor_geometry(monitor)
         geom = Rect(geom.x, geom.y, geom.width, geom.height)
 
         return area, geom
 
-    def get_docking_monitor(self):
-        screen = self.get_screen()
-        return screen.get_primary_monitor()
+    def get_docking_monitor(self, force_update=False):
+        if self._current_docking_monitor is None or force_update:
+            screen = self.get_screen()
+            docking_monitor = config.window.docking_monitor
+
+            monitor = int(docking_monitor)
+            if monitor < 100:
+                if monitor < 0 or \
+                   monitor >= screen.get_n_monitors():
+                    docking_monitor = DockingMonitor.PRIMARY
+
+            if docking_monitor == DockingMonitor.ACTIVE:
+                monitor = self._get_monitor_at_active_window(screen)
+            elif docking_monitor == DockingMonitor.PRIMARY:
+                monitor = screen.get_primary_monitor()
+        else:
+            monitor = self._current_docking_monitor
+        return monitor
+
+    @staticmethod
+    def _get_monitor_at_active_window(screen):
+        window = screen.get_active_window()
+        if window:
+            # Sometimes causes X error "BadDrawable" in Vivid for Francesco
+            Gdk.error_trap_push()
+            monitor = screen.get_monitor_at_window(window)
+            if Gdk.error_trap_pop():
+                monitor = screen.get_primary_monitor()
+        else:
+            monitor = screen.get_primary_monitor()
+        return monitor
 
     def reset_monitor_workarea(self):
         self._monitor_workarea = {}
 
-    def update_monitor_workarea(self):
+    def update_monitor_workarea(self, monitor):
         """
         Save the workarea, so we don't have to
         check all the time if our strut is already installed.
         """
-        mon = self.get_docking_monitor()
-        area = self.get_monitor_workarea()
-        self._monitor_workarea[mon] = area
+        area = self.get_monitor_workarea(monitor)
+        self._monitor_workarea[monitor] = area
         return area
 
-    def get_monitor_workarea(self):
+    def get_monitor_workarea(self, monitor):
         screen = self.get_screen()
-        mon = self.get_docking_monitor()
-        area = screen.get_monitor_workarea(mon)
+        area = screen.get_monitor_workarea(monitor)
         area = Rect(area.x, area.y, area.width, area.height)
         return area
 
@@ -1350,6 +1411,11 @@ class KbdWindow(KbdWindowBase, WindowRectPersist, Gtk.Window):
                 win.move_to_current_desktop()
                 self._moved_desktop_switch_count = self._desktop_switch_count
 
+    def raise_to_top(self):
+        win = self.get_window()
+        if win and self.is_visible():
+            win.raise_()
+
 
 class KbdPlugWindow(KbdWindowBase, Gtk.Plug):
     def __init__(self, keyboard_widget, icp = None):
