@@ -18,8 +18,7 @@ from Onboard.KeyGtk         import Key
 from Onboard.KeyCommon      import LOD
 from Onboard.TouchHandles   import TouchHandles
 from Onboard.LayoutView     import LayoutView
-from Onboard.AutoShow       import AutoShow
-from Onboard.utils          import Rect, Timer, FadeTimer
+from Onboard.utils          import Rect, Timer, FadeTimer, escape_markup
 from Onboard.definitions    import Handle
 from Onboard.WindowUtils    import WindowManipulator, \
                                    canvas_to_root_window_rect, \
@@ -84,7 +83,7 @@ class InactivityTimer(Timer):
     """
     Waits for the inactivity delay and transitions between
     active and inactive state.
-    Inactivity here means, the pointer has left the keyboard window
+    Inactivity here means, the pointer has left the keyboard window.
     """
     _keyboard = None
     _active = False
@@ -238,9 +237,6 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
 
         self.inactivity_timer = InactivityTimer(self)
 
-        self.auto_show = AutoShow(self)
-        self.auto_show.enable(config.is_auto_show_enabled())
-
         self.touch_handles = TouchHandles()
         self.touch_handles_hide_timer = Timer()
         self.touch_handles_fade = FadeTimer()
@@ -260,6 +256,7 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         self._configure_timer = Timer()
 
         self._language_menu = LanguageMenu(self)
+        self._suggestion_menu = SuggestionMenu(self)
 
         #self.set_double_buffered(False)
         self.set_app_paintable(True)
@@ -291,7 +288,6 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         self.inactivity_timer.stop()
         self._long_press_timer.stop()
         self._auto_release_timer.stop()
-        self.auto_show.cleanup()
         self.stop_click_polling()
         self._configure_timer.stop()
         self.close_key_popup()
@@ -311,7 +307,7 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         win = self.get_kbd_window()
         if win:
             self.touch_handles.set_window(win)
-            self.update_resize_handles()
+            self.update_window_handles()
 
     def set_startup_visibility(self):
         win = self.get_kbd_window()
@@ -423,7 +419,7 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
 
         return rect
 
-    def update_resize_handles(self):
+    def update_window_handles(self):
         """ Tell WindowManipulator about the active resize handles """
         docking = config.is_docking_enabled()
 
@@ -435,15 +431,6 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         self.touch_handles.set_active_handles(self._get_active_drag_handles(True))
         self.touch_handles.lock_x_axis(docking)
 
-    def update_auto_show(self):
-        """
-        Turn on/off auto-show in response to user action (preferences)
-        and show/hide the window accordingly.
-        """
-        enable = config.is_auto_show_enabled()
-        self.auto_show.enable(enable)
-        self.auto_show.show_keyboard(not enable)
-
     def update_transparency(self):
         """
         Updates transparencies in response to user action.
@@ -490,7 +477,7 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
 
         # stop reposition updates when we're hiding anyway
         if win and not visible:
-            win.stop_auto_position()
+            win.stop_auto_positioning()
 
         if config.is_docking_enabled():
             if slide_duration is None:
@@ -499,19 +486,32 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
             opacity_visible = True
 
             if win:
-                if visible:
-                    begin_rect = win.get_hidden_rect()
-                    end_rect = win.get_visible_rect()
-                else:
-                    begin_rect = win.get_rect()
-                    end_rect = win.get_docking_hideout_rect()
-                state.y.value = begin_rect.y
-                y             = end_rect.y
-                state.x.value = begin_rect.x
-                x             = end_rect.x
+                visible_before = win.is_visible()
+                visible_later  = visible
+                hideout_old_mon = win.get_docking_hideout_rect()
+                mon_changed = win.update_docking_monitor()
+                hideout_new_mon = win.get_docking_hideout_rect() \
+                                if mon_changed else hideout_old_mon
 
-                result |= self._init_transition(state.x, x, slide_duration)
-                result |= self._init_transition(state.y, y, slide_duration)
+                # Only position here if visibility or the active monitor
+                # changed. Leave it to auto_position to move the keyboard
+                # while it is visible, i.e. not being hidden or shown.
+                if visible_before != visible_later or \
+                   mon_changed:
+                    if visible:
+                        begin_rect = hideout_new_mon
+                        end_rect = win.get_visible_rect()
+                    else:
+                        begin_rect = win.get_rect()
+                        end_rect = hideout_old_mon
+
+                    state.y.value = begin_rect.y
+                    y             = end_rect.y
+                    state.x.value = begin_rect.x
+                    x             = end_rect.x
+
+                    result |= self._init_transition(state.x, x, slide_duration)
+                    result |= self._init_transition(state.y, y, slide_duration)
         else:
             opacity_visible  = visible
 
@@ -529,6 +529,10 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         return result
 
     def transition_active_to(self, active, duration = None):
+        """
+        Transition active state for inactivity timer.
+        This ramps up/down the window opacity.
+        """
         # not in xembed mode
         if config.xid_mode:
             return False
@@ -651,10 +655,6 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
 
         return not done
 
-    def toggle_visible(self):
-        """ main method to show/hide onboard manually """
-        self.set_visible(not self.is_visible())
-
     def is_visible(self):
         """ is the keyboard window currently visible? """
         window = self.get_kbd_window()
@@ -662,7 +662,6 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
 
     def set_visible(self, visible):
         """ main method to show/hide onboard manually """
-        self.lock_auto_show_visible(visible)  # pause auto show
         self.transition_visible_to(visible, 0.0)
 
         # briefly present the window
@@ -671,29 +670,24 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
             self.inactivity_timer.begin_transition(False)
 
         self.commit_transition()
-        win = self.get_kbd_window()
 
-    def lock_auto_show_visible(self, visible):
-        """
-        If the user unhides onboard, don't auto-hide it until
-        he manually hides it again.
-        """
-        if config.is_auto_show_enabled():
-            self.auto_show.lock_visible(visible)
+    def raise_to_top(self):
+        """ Raise the toplevel parent to top of the z-order. """
+        window = self.get_kbd_window()
+        if window:
+            window.raise_to_top()
 
-    def freeze_auto_show(self, thaw_time = None):
-        """
-        Stop both, hiding and showing.
-        """
-        if config.is_auto_show_enabled():
-            self.auto_show.freeze(thaw_time)
+    def auto_position(self):
+        """ auto-show, start repositioning """
+        window = self.get_kbd_window()
+        if window:
+            window.auto_position()
 
-    def thaw_auto_show(self, thaw_time = None):
-        """
-        Reenable both, hiding and showing.
-        """
-        if config.is_auto_show_enabled():
-            self.auto_show.thaw(thaw_time)
+    def stop_auto_positioning(self):
+        """ auto-show, stop all further repositioning attempts """
+        window = self.get_kbd_window()
+        if window:
+            window.stop_auto_positioning()
 
     def start_click_polling(self):
         if self.keyboard.has_latched_sticky_keys() or \
@@ -1056,7 +1050,8 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
     def on_drag_gesture_begin(self, num_touches):
         self.stop_long_press()
 
-        if num_touches and \
+        if Handle.MOVE in self.get_drag_handles() and \
+           num_touches and \
            not self.is_drag_initiated():
             self.show_touch_handles()
             self.start_move_window()
@@ -1268,7 +1263,7 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
             return 0.0
         co = self.get_kbd_window().get_orientation_config_object()
         if config.is_dock_expanded(co):
-            return 0.5
+            return 2.0
         if config.window.transparent_background:
             return 3.0
         return config.UNDECORATED_FRAME_WIDTH
@@ -1291,7 +1286,7 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
 
             if not all_handles:
                 # filter through handles enabled in config
-                config_handles = config.window.resize_handles
+                config_handles = config.window.window_handles
                 handles = tuple(set(handles).intersection(set(config_handles)))
 
         return handles
@@ -1406,19 +1401,32 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         root_rect = canvas_to_root_window_rect(self, r)
         popup.position_at(root_rect.x + root_rect.w * 0.5,
                          root_rect.y, 0.5, 1.0)
-        popup.show()
+        popup.show_all()
         return popup
 
     def show_snippets_dialog(self, snippet_id):
         """ Show dialog for creating a new snippet """
 
+        label, text = config.snippets.get(snippet_id, (None, None))
+        if snippet_id in config.snippets:
+            # Title of the snippets dialog for existing snippets
+            title = _format("Edit snippet #{}", snippet_id)
+            message = ""
+        else:
+            # Title of the snippets dialog for new snippets
+            title = _("New snippet")
+            # Message in the snippets dialog for new snippets
+            message = _format("Enter a new snippet for button #{}:", snippet_id)
+
         # turn off AT-SPI listeners to prevent D-BUS deadlocks (Quantal).
         self.keyboard.on_focusable_gui_opening()
 
-        dialog = Gtk.Dialog(title=_("New snippet"),
+        dialog = Gtk.Dialog(title=title,
                            transient_for=self.get_toplevel(),
                            flags=0)
-        dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
+        # Translators: cancel button of the snippets dialog. It used to
+        # be stock item STOCK_CANCEL until Gtk 3.10 deprecated those.
+        dialog.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
         dialog.add_button(_("_Save snippet"), Gtk.ResponseType.OK)
 
         # Don't hide dialog behind the keyboard in force-to-top mode.
@@ -1431,12 +1439,14 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
                       spacing=12, border_width=5)
         dialog.get_content_area().add(box)
 
-        msg = Gtk.Label(label=_("Enter a new snippet for this button:"),
-                        xalign=0.0)
-        box.add(msg)
+        if message:
+            msg_label = Gtk.Label(label=message, xalign=0.0)
+            box.add(msg_label)
 
         label_entry = Gtk.Entry(hexpand=True)
-        text_entry  = Gtk.Entry(hexpand=True, activates_default = True)
+        text_entry  = Gtk.Entry(hexpand=True,
+                                activates_default = True,
+                                width_chars=35)
         label_label = Gtk.Label(label=_("_Button label:"),
                                 xalign=0.0,
                                 use_underline=True,
@@ -1453,10 +1463,20 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
         grid.attach(text_entry, 1, 1, 1, 1)
         box.add(grid)
 
+        # Init entries, mainly the label for the case when text is empty.
+        label, text = config.snippets.get(snippet_id, (None, None))
+        if label:
+            label_entry.set_text(label)
+        if text:
+            text_entry.set_text(text)
+
+        if label and not text:
+            text_entry.grab_focus()
+        else:
+            label_entry.grab_focus()
+
         dialog.connect("response", self._on_snippet_dialog_response, \
                        snippet_id, label_entry, text_entry)
-
-        label_entry.grab_focus()
         dialog.show_all()
 
     def _on_snippet_dialog_response(self, dialog, response, snippet_id, \
@@ -1484,15 +1504,19 @@ class KeyboardWidget(Gtk.DrawingArea, WindowManipulator,
     def is_language_menu_showing(self):
         return self._language_menu.is_showing()
 
+    def show_prediction_menu(self, key, button, closure = None):
+        self._suggestion_menu.popup(key, button, closure)
 
-class LanguageMenu:
-    """ Popup menu for the language button """
+
+class KeyMenu:
+    """ Popup menu for keys """
 
     def __init__(self, keyboard_widget):
         self._keyboard_widget = keyboard_widget
         self._keyboard = self._keyboard_widget.keyboard
         self._menu = None
         self._closure = None
+        self._x_align = 0.0  # horizontal alignment of the menu position
 
     def is_showing(self):
         return not self._menu is None
@@ -1501,11 +1525,45 @@ class LanguageMenu:
         self._closure = closure
         self._keyboard.on_focusable_gui_opening()
 
+        menu = self.create_menu(key, button)
+        self._menu = menu
+
+        menu.connect("unmap", self._on_menu_unmap)
+        menu.show_all()
+        menu.popup(None, None, self._menu_positioning_func,
+                   key, button, Gtk.get_current_event_time())
+
+    def create_menu(self, key, button):
+        """ Overload this in derived class """
+        raise NotImplementedError()
+
+    def _on_menu_unmap(self, menu):
+        Timer(0.5, self._keyboard.on_focusable_gui_closed)
+        self._menu = None
+        if self._closure:
+            self._closure()
+
+    def _menu_positioning_func(self, menu, key):
+        r = self._keyboard_widget.get_key_screen_rect(key)
+        menu_size  = (menu.get_allocated_width(),
+                      menu.get_allocated_width())
+        x, y = self.get_menu_position(r, menu_size)
+        return x, y, False
+
+    def get_menu_position(self, rkey, menu_size):
+        return rkey.left() + (rkey.w - menu_size[0]) * self._x_align, \
+               rkey.bottom()
+
+
+class LanguageMenu(KeyMenu):
+    """ Popup menu for the language button """
+
+    def create_menu(self, key, button):
         keyboard = self._keyboard
         languagedb = keyboard._languagedb
 
         active_lang_id = keyboard.get_active_lang_id()
-        system_lang_id = keyboard.get_system_default_lang_id()
+        system_lang_id = config.get_system_default_lang_id()
 
         lang_ids = set(languagedb.get_language_ids())
         if system_lang_id in lang_ids:
@@ -1522,9 +1580,6 @@ class LanguageMenu:
             if name:
                 other_langs.append((name, lang_id))
 
-        def add_system_indicator(name):
-            return name + _(" - System Language")
-
         # language sub menu
         lang_menu = Gtk.Menu()
         for name, lang_id in sorted(other_langs):
@@ -1555,40 +1610,19 @@ class LanguageMenu:
                 item.connect("activate", self._on_language_activated, lang_id)
                 menu.append(item)
 
-        if other_langs:
-            if mru_lang_ids:
-                item = Gtk.MenuItem.new_with_mnemonic(_("Other _Languages"))
-            else:
-                item = Gtk.MenuItem.new_with_mnemonic(_("_Languages"))
-            item.set_submenu(lang_menu)
-            menu.append(item)
-
-        if lang_ids:
+        if mru_lang_ids:
             item = Gtk.SeparatorMenuItem.new()
             menu.append(item)
 
-        menu.connect("unmap", self._language_menu_unmap)
-        self._menu = menu
+        if other_langs:
+            item = Gtk.MenuItem.new_with_mnemonic(_("Other _Languages"))
+            item.set_submenu(lang_menu)
+            menu.append(item)
 
-        menu.show_all()
-
-        menu.popup(None, None, self._language_menu_positioning_func,
-                   key, button, Gtk.get_current_event_time())
-
-
-    def _language_menu_unmap(self, menu):
-        Timer(0.5, self._keyboard.on_focusable_gui_closed)
-        self._menu = None
-        if self._closure:
-            self._closure()
-
-    def _language_menu_positioning_func(self, menu, key):
-        r = self._keyboard_widget.get_key_screen_rect(key)
-        x = r.left()
-        return x, r.bottom(), True
+        return menu
 
     def _on_language_activated(self, menu, lang_id):
-        system_lang_id = self._keyboard.get_system_default_lang_id()
+        system_lang_id = config.get_system_default_lang_id()
         if lang_id == system_lang_id:
             lang_id = ""
         self._set_active_lang_id(lang_id)
@@ -1610,3 +1644,167 @@ class LanguageMenu:
         recent_languages = recent_languages[:max_recent_languages]
         config.typing_assistance.recent_languages = recent_languages
 
+
+class SuggestionMenu(KeyMenu):
+    """ Popup menu for word suggestion buttons """
+
+    def __init__(self, keyboard_widget):
+        KeyMenu.__init__(self, keyboard_widget)
+        self._x_align = 0.5
+
+    def create_menu(self, key, button):
+
+        self._choice_index = key.code
+
+        # popup menu
+        menu = Gtk.Menu()
+
+        item = Gtk.MenuItem.new_with_mnemonic(_("_Remove suggestion…"))
+        item.connect("activate", self._on_remove_suggestion)
+        menu.append(item)
+
+        return menu
+
+    def _on_remove_suggestion(self, menu_item):
+        keyboard = self._keyboard
+        suggestion, history = keyboard.get_prediction_choice_and_history( \
+                                                           self._choice_index)
+        history = history[-1:] # only single word history supported
+        dialog = RemoveSuggestionConfirmationDialog(
+                    self._keyboard_widget.get_kbd_window(),
+                    keyboard, suggestion, history)
+        context_length = dialog.run()
+        if context_length:
+            context = [suggestion]
+            if context_length == 2:
+                context = history[-1:] + context
+            keyboard.remove_prediction_context(context)
+
+            # Refresh word suggestions explicitely, the dialog
+            # disabled AT-SPI events.
+            keyboard.invalidate_context_ui()
+            keyboard.commit_ui_updates()
+
+    def get_menu_position(self, rkey, menu_size):
+        if menu_size[0] > rkey.w:
+            return rkey.left(), rkey.bottom()
+        else:
+            return super(SuggestionMenu, self).get_menu_position(
+                                                          rkey, menu_size)
+
+
+class RemoveSuggestionConfirmationDialog(Gtk.MessageDialog):
+    """ Confirm removal of a word suggestion """
+
+    def __init__(self, parent, keyboard, suggestion, history):
+        self._keyboard = keyboard
+        self._radio1 = None
+        self._radio2 = None
+
+        Gtk.MessageDialog.__init__(self,
+                                   message_type=Gtk.MessageType.QUESTION,
+                                   buttons=Gtk.ButtonsType.OK_CANCEL,
+                                   title=_("Onboard"))
+
+        if parent:
+            self.set_transient_for(parent)
+
+        # Don't hide dialog behind the keyboard in force-to-top mode.
+        #if config.is_force_to_top():
+        #    dialog.set_position(Gtk.WindowPosition.CENTER)
+
+        markup = "<big>" + _("Remove word suggestion") + "</big>"
+        markup = escape_markup(markup, preserve_tags=True)
+        self.set_markup(markup)
+
+        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
+
+        label = self._get_remove_context_length_2_label(suggestion, history)
+        self._radio2 = Gtk.RadioButton.new_with_label(None, label)
+        box.add(self._radio2)
+
+        self._radio1 = Gtk.RadioButton.new_with_label_from_widget(self._radio2,
+                       _format("Remove '{}' everywhere.", suggestion))
+        box.add(self._radio1)
+
+        if not history:
+            # This should rarly happen, if ever. Edited text usually
+            # start with the begin of text marker, so there exists a history
+            # even if the text appears to be empty.
+            self._radio2.set_sensitive(False)
+            self._radio1.set_active(True)
+
+        label = Gtk.Label(label=_("This will only affect learned suggestions."),
+                                   xalign=0.0)
+        box.add(label)
+
+        self.get_message_area().add(box)
+        self.show_all()
+
+    @staticmethod
+    def _get_remove_context_length_2_label(suggestion, history):
+        """
+        Label of radio button for remove context length 2.
+
+        Doctests:
+        >>> from Onboard.KeyboardWidget import RemoveSuggestionConfirmationDialog
+        >>> test = RemoveSuggestionConfirmationDialog._get_remove_context_length_2_label
+        >>> def _format(msgstr, *args, **kwargs):
+        ...     return msgstr.format(*args, **kwargs)
+        >>> import builtins
+        >>> builtins.__dict__['_format'] = _format
+
+        >>> test("word", [])
+        "Remove 'word' only where it occures at text begin."
+
+        >>> test("word", ["word2"])
+        "Remove 'word' only where it occures after 'word2'."
+
+        >>> test("word", ["word2", "word3"])
+        "Remove 'word' only where it occures after 'word2 word3'."
+
+        >>> test("word", ["<unk>"])
+        "Remove 'word' only where it occures after '<unk>'."
+
+        >>> test("word", ["<s>"])
+        "Remove 'word' only where it occures at sentence begin."
+
+        >>> test("word", ["<num>"])
+        "Remove 'word' only where it occures after numbers."
+        """
+        hist0 = history[-1] if history else None
+        if not hist0 or hist0.startswith("<bot:"):
+            label = _format("Remove '{}' only where it occures at text begin.",
+                            suggestion)
+        elif hist0 == "<s>":
+            label = _format("Remove '{}' only where it occures at sentence begin.",
+                            suggestion)
+        elif hist0 == "<num>":
+            label = _format("Remove '{}' only where it occures after numbers.",
+                            suggestion)
+        else:
+            label = _format("Remove '{}' only where it occures after '{}'.",
+                            suggestion, " ".join(history))
+        return label
+
+    def run(self):
+        keyboard = self._keyboard
+
+        if keyboard:
+            keyboard.on_focusable_gui_opening()
+
+        response = Gtk.Dialog.run(self)
+        self.destroy()
+
+        if keyboard:
+            keyboard.on_focusable_gui_closed()
+
+        # return length of the remove context
+        if response == Gtk.ResponseType.OK:
+            if self._radio2.get_active():
+                return 2
+            if self._radio1.get_active():
+                return 1
+        return 0
+
+
