@@ -3,19 +3,17 @@
 
 from __future__ import division, print_function, unicode_literals
 
-import os
 import time
 from math import pi
 
 import cairo
 from gi.repository         import Gtk, Gdk, GdkPixbuf
 
-from Onboard.utils         import Rect, Timer, \
+from Onboard.utils         import Rect, \
                                   roundrect_arc, roundrect_curve, \
-                                  gradient_line, brighten, timeit, \
+                                  gradient_line, brighten, \
                                   unicode_str
 from Onboard.WindowUtils   import get_monitor_dimensions
-from Onboard.definitions   import InputEventSourceEnum
 from Onboard.KeyGtk        import Key
 from Onboard.KeyCommon     import LOD
 
@@ -59,6 +57,7 @@ class LayoutView:
 
     def handle_realize_event(self):
         self.update_input_event_source()
+        self.update_touch_input_mode()
 
     def on_layout_loaded(self):
         """ Layout has been loaded. """
@@ -134,10 +133,16 @@ class LayoutView:
             self.keyboard.invalidate_canvas()
             self.keyboard.commit_ui_updates()
 
+    def is_visible(self):
+        return None
+
+    def set_visible(self, visible):
+        pass
+
     def toggle_visible(self):
         pass
 
-    def set_visible(self, visible):
+    def raise_to_top(self):
         pass
 
     def redraw(self, items = None, invalidate = True):
@@ -176,8 +181,10 @@ class LayoutView:
         pass
 
     def update_input_event_source(self):
-        use_gtk = config.keyboard.input_event_source == InputEventSourceEnum.GTK
-        self.register_input_events(True, use_gtk)
+        self.register_input_events(True, config.is_event_source_gtk())
+
+    def update_touch_input_mode(self):
+        self.set_touch_input_mode(config.keyboard.touch_input)
 
     def can_delay_sequence_begin(self, sequence):
         """
@@ -243,7 +250,7 @@ class LayoutView:
                (not self._starting_up or self._keys_pre_rendered)
 
     def draw(self, widget, context):
-        if not Gtk.cairo_should_draw_window(context, self.get_window()):
+        if not Gtk.cairo_should_draw_window(context, widget.get_window()):
             return
 
         lod = self._lod
@@ -301,7 +308,7 @@ class LayoutView:
         if config.xid_mode and \
            config.is_keep_aspect_ratio_enabled():
             if self.supports_alpha:
-                self._clear_xid_background(context)
+                self._clear_xembed_background(context)
                 transparent_bg = True
             else:
                 plain_bg = True
@@ -347,21 +354,21 @@ class LayoutView:
         context.paint()
         context.restore()
 
-    def _clear_xid_background(self, context):
+    def _clear_xembed_background(self, context):
         """ fill with plain layer 0 color; no alpha support required """
         rect = Rect(0, 0, self.get_allocated_width(),
                           self.get_allocated_height())
 
         # draw background image
         if config.get_xembed_background_image_enabled():
-            pixbuf = self._get_xid_background_image()
+            pixbuf = self._get_xembed_background_image()
             if pixbuf:
                 src_size = (pixbuf.get_width(), pixbuf.get_height())
                 x, y = 0, rect.bottom() - src_size[1]
                 Gdk.cairo_set_source_pixbuf(context, pixbuf, x, y)
                 context.paint()
 
-        # draw solid colored bar on top
+        # draw solid colored bar on top (with transparency, usually)
         rgba = config.get_xembed_background_rgba()
         if rgba is None:
             rgba = self.get_background_rgba()
@@ -370,7 +377,7 @@ class LayoutView:
         context.rectangle(*rect)
         context.fill()
 
-    def _get_xid_background_image(self):
+    def _get_xembed_background_image(self):
         """ load the desktop background image in Unity """
         try:
             pixbuf = self._xid_background_image
@@ -395,7 +402,7 @@ class LayoutView:
                                                 GdkPixbuf.InterpType.BILINEAR)
                 except Exception as ex: # private exception gi._glib.GError when
                                         # librsvg2-common wasn't installed
-                    _logger.error("_get_xid_background_image(): " + \
+                    _logger.error("_get_xembed_background_image(): " + \
                                 unicode_str(ex))
                     pixbuf = None
 
