@@ -83,41 +83,119 @@ class TextDomain:
         Get word separator to add after inserting a prediction choice.
 
         Doctests:
+
+        >>> from os.path import join
+
         # URL
         >>> d = DomainGenericText()
         >>> d.get_auto_separator("word http")
         ' '
         >>> d.get_auto_separator("word http://www")
         '.'
+        >>> d.get_auto_separator("word http://www.domain.org/path")
+        '/'
+        >>> d.get_auto_separator("word http://www.domain.org/path/document.ext")
+        ''
 
         # filename
         >>> d.get_auto_separator("/etc")
         '/'
         >>> from os.path import abspath, dirname
         >>> path = dirname(abspath("./onboard-defaults.conf.example"))
-        >>> d.get_auto_separator(path + "/onboard-defaults")
+        >>> d.get_auto_separator(join(path, "onboard-defaults"))
         '.'
-        >>> d.get_auto_separator(path + "/onboard-defaults.conf")
+        >>> d.get_auto_separator(join(path, "onboard-defaults.conf"))
         '.'
-        >>> d.get_auto_separator(path + "/onboard-defaults.conf.example")
+        >>> d.get_auto_separator(join(path, "onboard-defaults.conf.example"))
         ''
-        >>> d.get_auto_separator(path + "/onboard-defaults.conf.example.nexus7")
+        >>> d.get_auto_separator(join(path, "onboard-defaults.conf.example.nexus7"))
+        ' '
+
+        # filename in directory with spaces
+        >>> import tempfile
+        >>> td = tempfile.TemporaryDirectory(prefix="test onboard _")
+        >>> dir = td.name
+        >>> fn = os.path.join(dir, "onboard-defaults.conf.example")
+        >>> with open(fn, mode="w") as f: n = f.write("")
+        >>> d.get_auto_separator(join(dir, "onboard-defaults"))
+        '.'
+
+        # no false positives for valid filenames before the current token
+        >>> d.get_auto_separator(path + "/onboard-defaults no-file")
         ' '
         """
         separator = " "
 
-        # split at whitespace to catch whole URLs/file names
-        strings = context.split()
+        # Split at whitespace to catch whole URLs/file names and
+        # keep separators.
+        strings = re.split('(\s+)', context)
         if strings:
             string = strings[-1]
             if self._url_parser.is_maybe_url(string):
                 separator = self._url_parser.get_auto_separator(string)
-            elif self._url_parser._is_maybe_filename(string):
-                url = "file://" + string
-                separator = self._url_parser.get_auto_separator(url)
+            else:
+                fn = self._search_valid_file_name(strings)
+                if fn:
+                    string = fn
+                if self._url_parser._is_maybe_filename(string):
+                    url = "file://" + string
+                    separator = self._url_parser.get_auto_separator(url)
 
         return separator
 
+
+    def _search_valid_file_name(self, strings):
+        """
+        Search for a valid filename backwards across separators.
+
+        Doctests:
+        >>> import tempfile, re, os.path
+        >>>
+        >>> d = DomainGenericText()
+        >>>
+        >>> td = tempfile.TemporaryDirectory(prefix="test onboard _")
+        >>> dir = td.name
+        >>> fn1 = os.path.join(dir, "file")
+        >>> fn2 = os.path.join(dir, "file with many spaces")
+        >>> with open(fn1, mode="w") as f: n = f.write("")
+        >>> with open(fn2, mode="w") as f: n = f.write("")
+
+        # simple file in dir with spaces must return as filename
+        >>> strings = re.split('(\s+)', fn1)
+        >>> "/test onboard" in d._search_valid_file_name(strings)
+        True
+
+        # file with spaces in dir with spaces must return as filename
+        >>> strings = re.split('(\s+)', fn2)
+        >>> "/test onboard" in d._search_valid_file_name(strings)
+        True
+
+        # random string after a valid file must not be confused with a filename
+        >>> strings = re.split('(\s+)', fn2 + " no-file")
+        >>> d._search_valid_file_name(strings) is None
+        True
+        """
+        # Search backwards across spaces for an absolute filename.
+        max_sections = 16  # allow this many path sections (separators+tokens)
+        for i in range(min(max_sections, len(strings))):
+            fn = "".join(strings[-1-i:])
+            # Is it (part of) a valid absolute filename?
+            # Do least impact checks first.
+            if self._url_parser._is_maybe_filename(fn) and \
+               os.path.isabs(fn):
+
+                # Does a file or directory of this name exist?
+                if os.path.exists(fn):
+                    return fn
+
+                # Check if it is at least an incomplete filename of
+                # an existing file
+                files  = glob.glob(fn + "*")
+                if files:
+                    return fn
+
+        return None
+
     def grow_learning_span(self, text_span):
         """
         Grow span before learning to include e.g. whole URLs.
@@ -252,32 +330,57 @@ class DomainGenericText(TextDomain):
 
     def read_context(self, accessible):
         """ Extract prediction context from the accessible """
+
+        # get caret position from selection
+        selection = None
         try:
-            offset = accessible.get_caret_offset()
-            r = accessible.get_text_at_offset(offset,
+            sel = accessible.get_selection(0)
+            # Gtk-2 applications return 0,0 when there is no selection.
+            # Gtk-3 applications return caret positions in that case.
+            # LibreOffice Writer in Vivid initially returns -1,-1 when there
+            # is no selection, later the caret position.
+            start = sel.start_offset
+            end = sel.end_offset
+            if start > 0 and \
+               end > 0 and \
+               start <= end:
+                selection = (sel.start_offset, sel.end_offset)
+        except Exception as ex: # Private exception gi._glib.GErro
+            _logger.info("DomainGenericText.read_context(), selection: " \
+                         + unicode_str(ex))
+
+        # get text around the caret position
+        try:
+            if selection is None:
+                offset = accessible.get_caret_offset()
+                selection = (offset, offset)
+            r = accessible.get_text_at_offset(selection[0],
                                 Atspi.TextBoundaryType.LINE_START)
             count = accessible.get_character_count()
         except Exception as ex: # Private exception gi._glib.GError when
                                 # gedit became unresponsive.
-            _logger.info("DomainGenericText.read_context(): " \
+            _logger.info("DomainGenericText.read_context(), text: " \
                          + unicode_str(ex))
             return None
 
         line = unicode_str(r.content).replace("\n","")
-        line_caret = max(offset - r.start_offset, 0)
+        line_caret = max(selection[0] - r.start_offset, 0)
 
-        begin = max(offset - 256, 0)
-        end   = min(offset + 100, count)
+        begin = max(selection[0] - 256, 0)
+        end   = min(selection[0] + 100, count)
         text = Atspi.Text.get_text(accessible, begin, end)
 
         text = unicode_str(text)
 
-        caret_span = TextSpan(offset, 0, text, begin)
-        context = text[:offset - begin]
+        # Not all text may be available for large selections. We only need the
+        # part before the begin of the selection/caret.
+        selection_span = TextSpan(selection[0], selection[1]-selection[0],
+                              text, begin)
+        context = text[:selection[0] - begin]
         begin_of_text = begin == 0
         begin_of_text_offset = 0
 
-        return (context, line, line_caret, caret_span,
+        return (context, line, line_caret, selection_span,
                 begin_of_text, begin_of_text_offset)
 
     def can_spell_check(self):
@@ -331,61 +434,113 @@ class DomainTerminal(TextDomain):
                              + unicode_str(ex))
                 return None
 
-        # remove prompt from the current or previous lines
-        context, context_start, line, line_start, line_caret = \
-                self._read_after_prompt(accessible, offset)
-        if context_start:
+        context_lines, prompt_length, line, line_start, line_caret = \
+                             self._get_text_after_prompt(accessible, offset)
+
+        if prompt_length:
             begin_of_text = True
             begin_of_text_offset = line_start
         else:
             begin_of_text = False
             begin_of_text_offset = None
 
-        # remove newlines
-        context = context.replace("\n","")
+        context = "".join(context_lines)
+        before_line = "".join(context_lines[:-1])
+        selection_span = TextSpan(offset, 0,
+                                  before_line + line,
+                                  line_start-len(before_line))
 
-        #caret_span = TextSpan(offset, 0, text, begin)
-        caret_span = TextSpan(offset, 0, line, line_start)
-
-        result = (context, line, line_caret, caret_span,
+        result = (context, line, line_caret, selection_span,
                   begin_of_text, begin_of_text_offset)
         return result
 
-    def _read_after_prompt(self, accessible, offset):
-        r = accessible.get_text_at_offset(offset,
+    def _get_text_after_prompt(self, accessible, caret_offset):
+        """
+        Return text from the input area of the terminal after the prompt.
+
+        Doctests:
+        >>> class AtspiTextRangeMockup:
+        ...     pass
+        >>> class AccessibleMockup:
+        ...     def __init__(self, text, width):
+        ...         self._text = text
+        ...         self._width = width
+        ...     def get_text_at_offset(self, offset, boundary):
+        ...         line = offset // self._width
+        ...         lbegin = line * self._width
+        ...         r = AtspiTextRangeMockup()
+        ...         r.content = self._text[lbegin:lbegin+self._width]
+        ...         r.start_offset = lbegin
+        ...         return r
+        ...     def get_text_before_offset(self, offset, boundary):
+        ...         return self.get_text_at_offset(offset - self._width, boundary)
+
+        >>> d = DomainTerminal()
+
+        # Single line
+        >>> a = AccessibleMockup("abc$ ls /etc\\n", 15)
+        >>> d._get_text_after_prompt(a, 12)
+        (['ls /etc'], 5, 'ls /etc\\n', 5, 7)
+
+        # Two lines
+        >>> a = AccessibleMockup("abc$ ls /e"
+        ...                      "tc\\n", 10)
+        >>> d._get_text_after_prompt(a, 12)
+        (['ls /e', 'tc'], 5, 'tc\\n', 10, 2)
+
+        # Three lines: prompt not detected
+        # More that two lines are not supported. The probability of detecting
+        # "prompts" in random scrolling data rises with each additional line.
+        >>> a = AccessibleMockup("abc$ ls /e"
+        ...                      "tc/X11/xor"
+        ...                      "g.conf.d\\n", 10)
+        >>> d._get_text_after_prompt(a, 28)
+        (['tc/X11/xor', 'g.conf.d'], 0, 'g.conf.d\\n', 20, 8)
+
+        # Two lines with slash at the beginning of the second: detect vi
+        # search prompt. Not ideal, but vi is important too.
+        >>> a = AccessibleMockup("abc$ ls /etc"
+        ...                      "/X11\\n", 12)
+        >>> d._get_text_after_prompt(a, 16)
+        (['X11'], 1, 'X11\\n', 13, 3)
+
+        """
+
+        r = accessible.get_text_at_offset(caret_offset,
                             Atspi.TextBoundaryType.LINE_START)
-        line = unicode_str(r.content).replace("\n","")
+        line = unicode_str(r.content)
         line_start = r.start_offset
-        line_caret = offset - line_start
+        line_caret = caret_offset - line_start
 
         # remove prompt from the current or previous lines
-        context = ""
-        context_start = None
+        context_lines = []
+        prompt_length = None
         l = line[:line_caret]
         for i in range(2):
 
-            # blacklist matches? -> cancel whole context
+            # matching blacklisted prompt? -> cancel whole context
             if self._find_blacklisted_prompt(l):
-                context = ""
-                context_start = None
+                context_lines = []
+                prompt_length = None
                 break
 
-            context_start = self._find_prompt(l)
-            context = l[context_start:] + context
+            prompt_length = self._find_prompt(l)
+            context_lines.insert(0, l[prompt_length:])
             if i == 0:
-                line = line[context_start:] # cut prompt from input line
-                line_start  += context_start
-                line_caret -= context_start
-            if context_start:
+                line = line[prompt_length:] # cut prompt from input line
+                line_start += prompt_length
+                line_caret -= prompt_length
+            if prompt_length:
                 break
 
             # no prompt yet -> let context reach
             # across one more line break
-            r = accessible.get_text_before_offset(offset,
+            r = accessible.get_text_before_offset(caret_offset,
                                 Atspi.TextBoundaryType.LINE_START)
-            l = unicode_str(r.content).replace("\n","")
+            l = unicode_str(r.content)
 
-        result = (context, context_start, line, line_start, line_caret)
+        result = (context_lines, prompt_length,
+                  line, line_start, line_caret)
         return result
 
     def _find_prompt(self, context):
@@ -413,9 +568,9 @@ class DomainTerminal(TextDomain):
         # Only record (for learning) when there is a known prompt in sight.
         # Problem: learning won't happen for uncommon prompts, but less random
         # junk scrolling by should enter the user model in return.
-        context, context_start, line, line_start, line_caret = \
-                self._read_after_prompt(accessible, offset)
-        return bool(context_start)
+        context_lines, prompt_length, line, line_start, line_caret = \
+                self._get_text_after_prompt(accessible, offset)
+        return bool(prompt_length)
 
     def can_suggest_before_typing(self):
         """ Can give word suggestions before typing has started? """
@@ -533,18 +688,18 @@ class PartialURLParser:
         Is this maybe something looking like an URL?
 
         Doctests:
-        >>> d = PartialURLParser()
-        >>> d.is_maybe_url("http")
+        >>> p = PartialURLParser()
+        >>> p.is_maybe_url("http")
         False
-        >>> d.is_maybe_url("http:")
+        >>> p.is_maybe_url("http:")
         True
-        >>> d.is_maybe_url("http://www.domain.org")
+        >>> p.is_maybe_url("http://www.domain.org")
         True
-        >>> d.is_maybe_url("www.domain.org")
+        >>> p.is_maybe_url("www.domain.org")
         True
-        >>> d.is_maybe_url("www.domain")
+        >>> p.is_maybe_url("www.domain")
         False
-        >>> d.is_maybe_url("www")
+        >>> p.is_maybe_url("www")
         False
         """
         tokens = self.tokenize_url(context)
@@ -601,6 +756,8 @@ class PartialURLParser:
         >>> p.get_auto_separator("http://www.domain.co.uk")
         '/'
         >>> p.get_auto_separator("http://www.domain.org/home")
+        '/'
+        >>> p.get_auto_separator("http://www.domain.org/home/index.html")
         ''
         >>> p.get_auto_separator("mailto")
         ':'
@@ -686,6 +843,9 @@ class PartialURLParser:
             if context.startswith(file_scheme):
                 filename = context[len(file_scheme):]
                 separator = self._get_filename_separator(filename)
+            else:
+                if not last_septok == ".":
+                    separator = "/"
 
         if separator is None:
              separator = " " # may be entering search terms, keep space as default
