@@ -33,18 +33,14 @@ from gi.repository         import Gdk
 
 from Onboard.utils         import Timer, EventSource
 from Onboard.definitions   import TouchInputEnum
-from Onboard.XInput        import XIDeviceManager, XIEventType, XIEventMask
+from Onboard.XInput        import XIDeviceManager, XIEventType, XIEventMask, \
+                                  XIDeviceEventLogger
 
 ### Logging ###
 import logging
 _logger = logging.getLogger("TouchInput")
 ###############
 
-### Config Singleton ###
-from Onboard.Config import Config
-config = Config()
-########################
-
 BUTTON123_MASK = Gdk.ModifierType.BUTTON1_MASK | \
                  Gdk.ModifierType.BUTTON2_MASK | \
                  Gdk.ModifierType.BUTTON3_MASK
@@ -62,7 +58,7 @@ DRAG_GESTURE_THRESHOLD2 = 40**2  # square of the distance in pixels until
 # sequence id of core pointer events (single-touch/click events)
 POINTER_SEQUENCE = 0
 
-class InputEventSource(EventSource):
+class InputEventSource(EventSource, XIDeviceEventLogger):
     """
     Setup and handle GTK or XInput device events.
     """
@@ -71,6 +67,7 @@ class InputEventSource(EventSource):
         # There is only button-release to subscribe to currently,
         # as this is all CSButtonRemapper needs to detect the end of a click.
         EventSource.__init__(self, ["button-release"])
+        XIDeviceEventLogger.__init__(self)
 
         self._gtk_handler_ids = None
         self._device_manager = None
@@ -266,8 +263,8 @@ class InputEventSource(EventSource):
 
     def _select_xi_grab_events(self, select):
         """
-        Select events for the root window to simulate a pointer grab.
-        Only relevant when a drag was initiated, e.g. moving/resizing
+        Select root window events for simulating a pointer grab.
+        Only called when a drag was initiated, e.g. when moving/resizing
         the keyboard.
         """
         if select:
@@ -302,8 +299,12 @@ class InputEventSource(EventSource):
         event_type = event.xi_type
         device_id  = event.device_id
 
-        if _logger.isEnabledFor(logging.DEBUG):
-            self._log_event(event)
+        log = self._log_event_stub
+        if _logger.isEnabledFor(logging.DEBUG) and \
+            event_type != XIEventType.Motion and \
+            event_type != XIEventType.TouchUpdate:
+            self._log_device_event(event)
+            log = self.log_event
 
         # re-select devices on changes to the device hierarchy
         if event_type in XIEventType.HierarchyEvents or \
@@ -311,28 +312,40 @@ class InputEventSource(EventSource):
             self.select_xinput_devices()
             return
 
+        log("_on_device_event1")
+
+        if event_type == XIEventType.KeyPress or \
+           event_type == XIEventType.KeyRelease:
+            return
+
         # check device_id, discard duplicate and unknown events
         if event_type == XIEventType.Enter or \
            event_type == XIEventType.Leave:
 
+            log("_on_device_event2 {} {}", device_id, self._master_device_id)
             # enter/leave are only expected from the master device
             if not device_id == self._master_device_id:
+                log("_on_device_event3")
                 return
 
         else:
             # all other pointer/touch events have to come from slaves
+            log("_on_device_event4 {} {}", event.device_id, self._slave_device_ids)
             if not event.device_id in self._slave_device_ids:
+                log("_on_device_event5")
                 return
 
         # bail if the window isn't realized yet
         win = self.get_window()
         if not win:
+            log("_on_device_event6")
             return
 
         # scale coordinates in response to changes to
         # org.gnome.desktop.interface scaling-factor
         try:
             scale = win.get_scale_factor()  # from Gdk 3.10
+            log("_on_device_event7 {}", scale)
             if scale and scale != 1.0:
                 scale = 1.0 / scale
                 event.x = event.x * scale
@@ -348,14 +361,14 @@ class InputEventSource(EventSource):
         # -> Simulate pointer grab, select root events we can track even
         #    outside the keyboard window.
         # None of these problems are assumed to exist for touch devices.
+        log("_on_device_event8 {}", self._xi_grab_active)
         if self._xi_grab_active and \
            (event_type == XIEventType.Motion or \
             event_type == XIEventType.ButtonRelease):
             if not self._xi_grab_events_selected:
                 self._select_xi_grab_events(True)
 
-            #print(event_type, event.x_root, event.y_root)
-            #print(self._xi_grab_active, event_type, event.state, event.device_id, self._master_device_id, event.xid_event)
+            log("_on_device_event9")
 
             # We only get root window coordinates for root window events,
             # so convert them to our target window's coordinates.
@@ -370,8 +383,11 @@ class InputEventSource(EventSource):
             # one finger, touching anything in a long press popup must
             # not also affect the keyboard below.
             xid_event = event.xid_event
+            xid_win = win.get_xid()
+            log("_on_device_event10 {} {}", xid_event, xid_win)
             if xid_event != 0 and \
-                xid_event != win.get_xid():
+                xid_event != xid_win:
+                log("_on_device_event11")
                 return
 
         # Dispatch events
@@ -400,11 +416,11 @@ class InputEventSource(EventSource):
         elif event_type == XIEventType.Leave:
             self._on_leave_notify(self, event)
 
-    def _log_event(self, event):
+    def _log_device_event(self, event):
         win = self.get_window()
         if not event.xi_type in [ XIEventType.TouchUpdate,
                                   XIEventType.Motion]:
-            _logger.debug("Device event: dev_id={} src_id={} xi_type={} "
+            self.log_event("Device event: dev_id={} src_id={} xi_type={} "
                           "xid_event={}({}) x={} y={} x_root={} y_root={} "
                           "button={} state={} sequence={}"
                           "".format(event.device_id,
@@ -419,6 +435,17 @@ class InputEventSource(EventSource):
                                    )
                          )
 
+            device = event.get_source_device()
+            self.log_event("Source device: " + str(device))
+
+    @staticmethod
+    def log_event(msg, *args):
+        _logger.event(msg.format(*args))
+
+    @staticmethod
+    def _log_event_stub(msg, *args):
+        pass
+
 
 class TouchInput(InputEventSource):
     """
@@ -434,10 +461,11 @@ class TouchInput(InputEventSource):
         InputEventSource.__init__(self)
 
         self._input_sequences = {}
-        self._touch_events_enabled = self.is_touch_enabled()
-        self._multi_touch_enabled  = config.keyboard.touch_input == \
-                                     TouchInputEnum.MULTI
-        self._gestures_enabled     = self._touch_events_enabled
+
+        self._touch_events_enabled = False
+        self._multi_touch_enabled  = False
+        self._gestures_enabled     = False
+
         self._last_event_was_touch = False
         self._last_sequence_time = 0
 
@@ -449,8 +477,22 @@ class TouchInput(InputEventSource):
         self._num_tap_sequences = 0
         self._gesture_timer = Timer()
 
-    def is_touch_enabled(self):
-        return config.keyboard.touch_input != TouchInputEnum.NONE
+    def set_touch_input_mode(self, touch_input):
+        """ Call this to enable single/multi-touch """
+        self._touch_events_enabled = touch_input != TouchInputEnum.NONE
+        self._multi_touch_enabled  = touch_input == TouchInputEnum.MULTI
+        self._gestures_enabled     = self._touch_events_enabled
+        if self._device_manager:
+            self._device_manager.update_devices() # reset touch_active
+
+        _logger.debug("setting touch input mode {}: "
+                      "touch_events_enabled={}, "
+                      "multi_touch_enabled={}, "
+                      "gestures_enabled={}" \
+                      .format(touch_input,
+                              self._touch_events_enabled,
+                              self._multi_touch_enabled,
+                              self._gestures_enabled))
 
     def has_input_sequences(self):
         """ Are any clicks/touches still ongoing? """
@@ -460,20 +502,51 @@ class TouchInput(InputEventSource):
         """ Was there just a touch event? """
         return self._last_event_was_touch
 
-    def has_touch_source(self, event):
-        """ Was source device of event a touch screen? """
-        source_device = event.get_source_device()
-        source = source_device.get_source()
-        return source == Gdk.InputSource.TOUCHSCREEN
+    @staticmethod
+    def _get_event_source(event):
+        device = event.get_source_device()
+        return device.get_source()
+
+    @staticmethod
+    def _set_touch_active(event, active):
+        """ Mark source device as actively receiving touch events """
+        device = event.get_source_device()
+        device.touch_active = active
+
+    def _can_handle_pointer_event(self, event):
+        """
+        Rely on pointer events? True for non-touch devices
+        and wacom touch-screens with gestures enabled.
+        """
+        device = event.get_source_device()
+        source = device.get_source()
+
+        return not self._touch_events_enabled or \
+               source != Gdk.InputSource.TOUCHSCREEN or \
+               not device.touch_active
+
+    def _can_handle_touch_event(self, event):
+        """
+        Rely on touch events? True for touch devices
+        and wacom touch-screens with gestures disabled.
+        """
+        return not self._can_handle_pointer_event(event)
 
     def _on_button_press_event(self, widget, event):
-        if self._touch_events_enabled and \
-           self.has_touch_source(event):
-                return
+        self.log_event("_on_button_press_event1 {} {} {} ",
+                       self._touch_events_enabled,
+                       self._can_handle_pointer_event(event),
+                       self._get_event_source(event))
+
+        if not self._can_handle_pointer_event(event):
+            self.log_event("_on_button_press_event2 abort")
+            return
 
         # - Ignore double clicks (GDK_2BUTTON_PRESS),
         #   we're handling those ourselves.
         # - Ignore mouse wheel button events
+        self.log_event("_on_button_press_event3 {} {}",
+                       event.type, event.button)
         if event.type == Gdk.EventType.BUTTON_PRESS and \
            1 <= event.button <= 3:
             sequence = InputSequence()
@@ -481,10 +554,14 @@ class TouchInput(InputEventSource):
             sequence.primary = True
             self._last_event_was_touch = False
 
+            self.log_event("_on_button_press_event4")
             self._input_sequence_begin(sequence)
 
+        return True
+
     def _on_button_release_event(self, widget, event):
         sequence = self._input_sequences.get(POINTER_SEQUENCE)
+        self.log_event("_on_button_release_event", sequence)
         if not sequence is None:
             sequence.point      = (event.x, event.y)
             sequence.root_point = (event.x_root, event.y_root)
@@ -492,10 +569,11 @@ class TouchInput(InputEventSource):
 
             self._input_sequence_end(sequence)
 
+        return True
+
     def _on_motion_event(self, widget, event):
-        if self._touch_events_enabled and \
-           self.has_touch_source(event):
-                return
+        if not self._can_handle_pointer_event(event):
+            return
 
         sequence = self._input_sequences.get(POINTER_SEQUENCE)
         if sequence is None and \
@@ -509,21 +587,38 @@ class TouchInput(InputEventSource):
             self._last_event_was_touch = False
             self._input_sequence_update(sequence)
 
+        return True
+
     def _on_enter_notify(self, widget, event):
         self.on_enter_notify(widget, event)
+        return True
 
     def _on_leave_notify(self, widget, event):
         self.on_leave_notify(widget, event)
+        return True
 
     def _on_touch_event(self, widget, event):
-        if not self.has_touch_source(event):
+        self.log_event("_on_touch_event1 {}", self._get_event_source(event))
+
+        event_type = event.type
+
+        # Set source_device touch-active to block processing of pointer events.
+        # "touch-screens" that don't send touch events will keep having pointer
+        # events handled (Wacom devices with gestures enabled).
+        # This assumes that for devices that emit both touch and pointer
+        # events, the touch event comes first. Else there will be a dangling
+        # touch sequence. _discard_stuck_input_sequences would clean that up,
+        # but a key might get still get stuck in pressed state.
+        self._set_touch_active(event, True)
+
+        if not self._can_handle_touch_event(event):
+            self.log_event("_on_touch_event2 abort")
             return
 
-        touch = event.touch
+        touch = event.touch if hasattr(event, "touch") else event
         id = str(touch.sequence)
         self._last_event_was_touch = True
 
-        event_type = event.type
         if event_type == Gdk.EventType.TOUCH_BEGIN:
             sequence = InputSequence()
             sequence.init_from_touch_event(touch, id)
@@ -554,8 +649,11 @@ class TouchInput(InputEventSource):
                 sequence.time = event.get_time()
                 self._input_sequence_end(sequence)
 
+        return True
+
     def _input_sequence_begin(self, sequence):
         """ Button press/touch begin """
+        self.log_event("_input_sequence_begin1 {}", sequence)
         self._gesture_sequence_begin(sequence)
         first_sequence = len(self._input_sequences) == 0
 
@@ -594,6 +692,7 @@ class TouchInput(InputEventSource):
         return False
 
     def deliver_input_sequence_begin(self, sequence):
+        self.log_event("deliver_input_sequence_begin {}", sequence)
         self.on_input_sequence_begin(sequence)
         sequence.delivered = True
 
@@ -607,12 +706,14 @@ class TouchInput(InputEventSource):
 
     def _input_sequence_end(self, sequence):
         """ Button release/touch end """
+        self.log_event("_input_sequence_end1 {}", sequence)
         self._gesture_sequence_end(sequence)
         self._gesture_timer.finish()  # run delayed begin before end
         if sequence.id in self._input_sequences:
             del self._input_sequences[sequence.id]
 
             if sequence.delivered:
+                self.log_event("_input_sequence_end2 {}", sequence)
                 self.on_input_sequence_end(sequence)
 
         if self._input_sequences:
@@ -791,4 +892,20 @@ class InputSequence:
     def __repr__(self):
         return "{}({})".format(type(self).__name__,
                                repr(self.id))
+    def __str__(self):
+        return "{}(id={} point=({:.2f}, {:.2f}) root_point=({:.2f}, {:.2f}) " \
+               "button={} state={} event_type={} time={} primary={} delivered={} " \
+               "active_key={})" \
+                .format(type(self).__name__,
+                        self.id,
+                        self.point[0], self.point[1],
+                        self.root_point[0], self.root_point[1],
+                        self.button,
+                        self.state,
+                        self.event_type,
+                        self.time,
+                        self.primary,
+                        self.delivered,
+                        self.active_key,
+                       )
 
