@@ -20,6 +20,7 @@
 from __future__ import division, print_function, unicode_literals
 
 import sys
+import copy
 
 from gi.repository import Gdk
 
@@ -121,7 +122,7 @@ class XIDeviceManager(EventSource):
     XInput device manager singleton.
     """
 
-    blacklist = ("Power Button")
+    blacklist = ("Virtual core XTEST keyboard", "Power Button")
     last_device_blacklist = ("Virtual core XTEST pointer")
 
     def __new__(cls, *args, **kwargs):
@@ -180,6 +181,12 @@ class XIDeviceManager(EventSource):
         device_id = self._osk_devices.get_client_pointer()
         return self.lookup_device_id(device_id)
 
+    def get_client_keyboard(self):
+        """ Return client pointer device """
+        client_pointer = self.get_client_pointer()
+        device_id = client_pointer.attachment
+        return self.lookup_device_id(device_id)
+
     def get_devices(self):
         return self._devices.values()
 
@@ -194,7 +201,7 @@ class XIDeviceManager(EventSource):
         client_pointer = self.get_client_pointer()
         return [device for device in self.get_pointer_devices() \
                 if not device.is_master() and \
-                   device.master == client_pointer.id]
+                   device.attachment == client_pointer.id]
 
     def get_client_pointer_attached_slaves(self):
         """
@@ -205,6 +212,30 @@ class XIDeviceManager(EventSource):
                                          not self.is_grabbed(d)]
         return devices
 
+    def get_keyboard_devices(self):
+        return [device for device in self._devices.values() \
+                if device.is_keyboard()]
+
+    def get_client_keyboard_slaves(self):
+        """
+        All slaves of the keyboard paired with the client pointer,
+        with and without device grabs.
+        """
+        client_keyboard = self.get_client_keyboard()
+        return [device for device in self.get_keyboard_devices() \
+                if not device.is_master() and \
+                   device.attachment == client_keyboard.id]
+
+    def get_client_keyboard_attached_slaves(self):
+        """
+        Slaves that are currently attached to the keyboard paired
+        with the client pointer.
+        """
+        devices = self.get_client_keyboard_slaves()
+        devices = [d for d in devices if not d.is_floating() and \
+                                         not self.is_grabbed(d)]
+        return devices
+
     def get_master_pointer_devices(self):
         return [device for device in self.get_pointer_devices() \
                 if device.is_master()]
@@ -220,14 +251,14 @@ class XIDeviceManager(EventSource):
                 device.name,
                 device.id,
                 device.use,
-                device.master,
+                device.attachment,
                 device.enabled,
                 device.vendor,
                 device.product,
-                touch_mode,
+                device.touch_mode,
             ) = info
             device.source = XIDevice.classify_source(device.name, device.use,
-                                                      touch_mode)
+                                                     device.touch_mode)
 
             if sys.version_info.major == 2:
                 device.name = unicode_str(device.name)
@@ -334,6 +365,12 @@ class XIDeviceManager(EventSource):
             source_device = None
         event.set_source_device(source_device)
 
+        ## debug, simulate touch-screen
+        if 0 and \
+           device_id == 11:
+            if not self._disguise_as_touch_event(event, False):
+                return
+
         # remember recently used device ids for CSFloatingSlave
         if not source_id in self._last_device_blacklist_ids:
             if event_type == XIEventType.Motion:
@@ -343,7 +380,48 @@ class XIDeviceManager(EventSource):
                 self._last_click_device_id = source_id
 
         # forward the event to all listeners
-        self.emit("device-event", event)
+        for callback in self._callbacks["device-event"]:
+            # Copy event to isolate callbacks from each other (LP: 1421840)
+            ev = copy.copy(event)
+            callback(ev)
+
+    def _disguise_as_touch_event(self, event, wacom_mode=False):
+        """
+        Disguise a pointer event as touch event.
+        For debugging purposes only.
+
+        Set wacom_mode=True to simulate pointer events coming
+        from a "touch screen" device that only sends pointer events.
+        This is the default case for wacom touch screens due to
+        gestures being enabled, see LP #1297692.
+        """
+        device = self.lookup_device_id(event.source_id)
+        device.name         = "Touch-Screen"
+        device.source       = Gdk.InputSource.TOUCHSCREEN
+        device.touch_mode   = XITouchMode.DirectTouch
+
+        if event.xi_type == XIEventType.Motion and \
+           not event.state & (Gdk.ModifierType.BUTTON1_MASK | \
+                                Gdk.ModifierType.BUTTON2_MASK | \
+                                Gdk.ModifierType.BUTTON3_MASK):
+            return False  # discard event
+
+        if not wacom_mode:
+            if event.xi_type == XIEventType.ButtonPress:
+                event.xi_type = XIEventType.TouchBegin
+                event.type = Gdk.EventType.TOUCH_BEGIN
+            if event.xi_type == XIEventType.ButtonRelease:
+                event.xi_type = XIEventType.TouchEnd
+                event.type = Gdk.EventType.TOUCH_END
+            if event.xi_type == XIEventType.Motion:
+                event.xi_type = XIEventType.TouchUpdate
+                event.type = Gdk.EventType.TOUCH_UPDATE
+
+            event.sequence = 10  # single touch only
+
+        event.button = 1
+
+        return True  # allow event
 
 
 class XIDevice(object):
@@ -353,23 +431,44 @@ class XIDevice(object):
     name         = None
     id           = None
     use          = None
-    master       = None
+    attachment   = None  # master for slaves, paired master for masters
     enabled      = None
     vendor       = None
     product      = None
     source       = None
+    touch_mode   = None
+
+    touch_active = False  # If False, only pointer events are considered.
+                          # Wacom devices with enabled gestures never
+                          # become touch-active, i.e. they don't generate
+                          # touch events.
 
     _device_manager = None
 
-    def __str__(self):
-        return "{}(id={}, master={}, name={}, source={} )" \
+    def __repr__(self):
+        return "{}(id={}, attachment={}, name={}, source={} )" \
                 .format(type(self).__name__,
                         repr(self.id),
-                        repr(self.master),
+                        repr(self.attachment),
                         repr(self.name),
                         repr(self.source),
                        )
 
+    def __str__(self):
+        return "{}(id={} attachment={} use={} touch_mode={} source={} name={} " \
+               "vendor=0x{:04x} product=0x{:04x} enabled={})" \
+                .format(type(self).__name__,
+                        self.id,
+                        self.attachment,
+                        self.use,
+                        self.touch_mode,
+                        self.get_source().value_name,
+                        self.name,
+                        self.vendor,
+                        self.product,
+                        self.enabled,
+                       )
+
     def get_source(self):
         """
         Return Gdk.InputSource for compatibility with Gtk event handling.
@@ -425,6 +524,13 @@ class XIDevice(object):
         return self.use == XIDeviceType.MasterPointer or \
                self.use == XIDeviceType.SlavePointer
 
+    def is_keyboard(self):
+        """
+        Is this device a keyboard?
+        """
+        return self.use == XIDeviceType.MasterKeyboard or \
+               self.use == XIDeviceType.SlaveKeyboard
+
     def is_floating(self):
         """
         Is this device detached?
@@ -441,3 +547,47 @@ class XIDevice(object):
                                            self.product,
                                            self.use)
 
+class XIDeviceEventLogger:
+    """
+    Facilities for logging device events.
+    Has little overhead when logging is disabled.
+    """
+
+    def __init__(self):
+        if not _logger.isEnabledFor(logging.DEBUG):
+            self.log_event = self._log_event_stub
+
+    def get_window(self):
+        return None
+
+    def _log_device_event(self, event):
+        win = self.get_window()
+        if not event.xi_type in [ XIEventType.TouchUpdate,
+                                  XIEventType.Motion]:
+            self.log_event("Device event: dev_id={} src_id={} xi_type={} "
+                          "xid_event={}({}) x={} y={} x_root={} y_root={} "
+                          "button={} state={} sequence={}"
+                          "".format(event.device_id,
+                                    event.source_id,
+                                    event.xi_type,
+                                    event.xid_event,
+                                    win.get_xid() if win else 0,
+                                    event.x, event.y,
+                                    event.x_root, event.y_root,
+                                    event.button, event.state,
+                                    event.sequence,
+                                   )
+                         )
+
+            device = event.get_source_device()
+            self.log_event("Source device: " + str(device))
+
+    @staticmethod
+    def log_event(msg, *args):
+        _logger.event(msg.format(*args))
+
+    @staticmethod
+    def _log_event_stub(msg, *args):
+        pass
+
+
