@@ -25,6 +25,9 @@ Author: marmuta <marmvta@gmail.com>
 
 #include "lm.h"
 
+#define HONOR_REMOVED_NODES true
+
+
 #pragma pack(2)
 
 //------------------------------------------------------------------------
@@ -210,7 +213,23 @@ class BeforeLastNode : public TBASE
             return lo;
         }
 
-        int get_N1prx() {return children.size();}  // assumes all have counts>=1
+        int get_N1prx()
+        {
+            // Take removed nodes into account (count==0)
+            int n = 0;
+            if (HONOR_REMOVED_NODES)  // any removed nodes in the model?
+            {
+                for (int i=0; i<children.size(); i++)
+                    if (children[i].get_count() > 0)
+                        n++;
+            }
+            else
+            {
+                n = children.size();  // assumes all have counts>=1
+            }
+
+            return n;
+        }
 
         int sum_child_counts()
         {
@@ -285,13 +304,23 @@ class TrieNode : public TBASE
 
         int get_N1prx()
         {
-            int n = children.size();  // assumes all children have counts > 0
+            int n = 0;
+            if (HONOR_REMOVED_NODES)  // any removed nodes in the model?
+            {
+                for (int i=0; i<(int)children.size(); i++)
+                    if (children[i]->get_count() > 0)
+                        n++;
+            }
+            else
+            {
+                n = children.size();  // assumes all children have counts > 0
 
-            // Unigrams <unk>, <s>,... may be empty initially. Don't count them
-            // or predictions for small models won't sum close to 1.0
-            for (int i=0; i<n && i<NUM_CONTROL_WORDS; i++)
-                if (children[0]->get_count() == 0)
-                    n--;
+                // Unigrams <unk>, <s>,... may be empty initially. Don't count them
+                // or predictions for small models won't sum close to 1.0
+                for (int i=0; i<n && i<NUM_CONTROL_WORDS; i++)
+                    if (children[0]->get_count() == 0)
+                        n--;
+            }
             return n;
         }
 
@@ -338,6 +367,19 @@ class NGramTrie : public TNODE
                 }
 
                 void operator++(int unused) // postfix operator
+                {
+                    BaseNode* node;
+                    for(;;)
+                    {
+                        node = next();
+                        // skip removed nodes, i.e. nodes with count==0
+                        if (node == NULL || node->count != 0)
+                            break;
+                    }
+                }
+
+                // next for all nodes, including deleted ones with count==0
+                BaseNode* next()
                 {
                     // preorder traversal with shallow stack
                     // nodes stack: path to node
@@ -351,7 +393,7 @@ class NGramTrie : public TNODE
                         nodes.pop_back();
                         indexes.pop_back();
                         if (nodes.empty())
-                            return;
+                            return NULL;
 
                         node = nodes.back();
                         index = ++indexes.back();
@@ -362,6 +404,7 @@ class NGramTrie : public TNODE
                     nodes.push_back(node);
                     indexes.push_back(0);
                     //printf ("pushed %d %d %d\n", nodes.back()->word_id, index, indexes.back());
+                    return node;
                 }
 
                 void get_ngram(std::vector<WordId>& ngram)
@@ -419,7 +462,24 @@ class NGramTrie : public TNODE
                                  int increment)
         {
             total_ngrams[n-1] += increment;
+
+            // Adding n-gram?
+            if (node->count == 0 && increment > 0)
+                num_ngrams[n-1]++;
+
             node->count += increment;
+
+            // Removing n-gram?
+            if (node->count <= 0 && increment < 0)
+            {
+                num_ngrams[n-1]--;
+
+                // Control words must not be removed.
+                if (n == 1 && wids[0] < NUM_CONTROL_WORDS)
+                {
+                    node->count = 1;
+                }
+            }
             return node->count;
         }
 
@@ -438,13 +498,17 @@ class NGramTrie : public TNODE
                                   int num_word_types,
                                   const std::vector<double>& Ds);
 
-        // get number of unique ngrams
+        // Get number of unique ngrams per level, excluding removed ones
+        // with count==0.
         int get_num_ngrams(int level) { return num_ngrams[level]; }
 
-        // get total number of all ngram occurences
+        // Get total number of all ngram occurences per level.
         int get_total_ngrams(int level) { return total_ngrams[level]; }
 
-        // get number of occurences of a specific ngram
+        // Number of distinct words excluding removed ones with count=0.
+        virtual int get_num_word_types() {return get_num_ngrams(0);}
+
+        // Get number of occurences of a specific ngram.
         int get_ngram_count(const std::vector<WordId>& wids)
         {
             BaseNode* node = get_node(wids);
@@ -493,6 +557,8 @@ class NGramTrie : public TNODE
             return static_cast<TNODE*>(parent)->children[index];
         }
 
+        // Return the word ids of all direct child nodes,
+        // excluding removed n-grams, i.g. count == 0.
         void get_child_wordids(const std::vector<WordId>& wids,
                           std::vector<WordId>& child_wids)
         {
@@ -504,7 +570,8 @@ class NGramTrie : public TNODE
                 for(int i=0; i<num_children; i++)
                 {
                     BaseNode* child = get_child_at(node, level, i);
-                    child_wids.push_back(child->word_id);
+                    if (child->count)
+                        child_wids.push_back(child->word_id);
                 }
             }
         }
@@ -597,7 +664,14 @@ class NGramTrie : public TNODE
 
     public:
         int order;
+
+        // Keep track of these counts to avoid
+        // traversing the tree for these numbers.
+        //
+        // Number of unique ngrams with count > 0, per level.
         std::vector<int> num_ngrams;
+
+        // Number of total occurences of all n-grams, per level.
         std::vector<int> total_ngrams;
 };
 
@@ -635,18 +709,30 @@ class DynamicModelBase : public NGramModel
         virtual void clear()
         {
             LanguageModel::clear();
+            assure_valid_control_words();
+        }
+
+        // Make sure control words exist as unigrams.
+        // They must have a count of at least 1. 0 means removed and
+        // it also throws off the normalization of witten-bell smoothing.
+        virtual void assure_valid_control_words()
+        {
+            const wchar_t* words[NUM_CONTROL_WORDS] =
+                        {L"<unk>", L"<s>", L"</s>", L"<num>"};
 
-            // Add entries for control words.
-            // Add them with a count of 1 as 0 throws off the normalization
-            // of witten-bell smoothing.
-            const wchar_t* words[] = {L"<unk>", L"<s>", L"</s>", L"<num>"};
             for (WordId i=0; i<ALEN(words); i++)
             {
-                count_ngram(words+i, 1, 1);
+                // Control words must have fixed positions at
+                // the very beginning of the dictionary.
                 assert(dictionary.word_to_id(words[i]) == i);
+
+                if (get_ngram_count(words+i, 1) <= 0)
+                    count_ngram(words+i, 1, 1);
+
             }
         }
 
+        virtual int  get_ngram_count(const wchar_t* const* ngram, int n) = 0;
         virtual void get_node_values(BaseNode* node, int level,
                                      std::vector<int>& values) = 0;
         virtual BaseNode* count_ngram(const wchar_t* const* ngram, int n,
@@ -659,16 +745,41 @@ class DynamicModelBase : public NGramModel
         virtual LMError save(const char* filename)
         {return save_arpac(filename);}
 
+        // Debug output, dump all n-grams.
+        virtual void dump()
+        {
+            std::vector<WordId> wids;
+            DynamicModelBase::ngrams_iter* it;
+            for (it = ngrams_begin(); ; (*it)++)
+            {
+                BaseNode* node = *(*it);
+                if (!node)
+                    break;
+
+                it->get_ngram(wids);
+
+                std::vector<int> values;
+                get_node_values(node, wids.size(), values);
+
+                unsigned i;
+                for (i=0; i<wids.size(); i++)
+                    printf("%ls ", dictionary.id_to_word(wids[i]));
+                for (i=0; i<values.size(); i++)
+                    printf("%d ", values[i]);
+                printf("\n");
+            }
+            printf("\n");
+        }
+
     protected:
         // temporary unigram, only used during loading
-        typedef struct 
+        typedef struct
         {
             std::wstring word;
             uint32_t count;
             uint32_t time;
         } Unigram;
-        virtual LMError
-                   set_unigrams(const std::vector<Unigram>& unigrams);
+        virtual LMError set_unigrams(const std::vector<Unigram>& unigrams);
 
         virtual LMError write_arpa_ngram(FILE* f,
                                        const BaseNode* node,
@@ -694,6 +805,9 @@ class DynamicModelBase : public NGramModel
         virtual int get_num_ngrams(int level) = 0;
         virtual void reserve_unigrams(int count) = 0;
 
+        // Number of distinct words excluding removed ones with count=0.
+        virtual int get_num_word_types() {return get_num_ngrams(0);}
+
 };
 
 
@@ -765,6 +879,38 @@ class _DynamicModel : public DynamicModelBase
             return smoothings;
         }
 
+        virtual void filter_candidates(const std::vector<WordId>& in,
+                                             std::vector<WordId>& out)
+        {
+            // filter out removed unigrams
+            int num_candidates = in.size();
+            out.reserve(num_candidates);
+            for (int i=0; i<num_candidates; i++)
+            {
+                WordId wid = in[i];
+                // can crash if is_model_valid() == false
+                BaseNode* node = ngrams.get_child_at(&ngrams, 0, wid);
+                if (node->get_count())
+                    out.push_back(wid);
+            }
+        }
+
+        // Plausibilty check befor predict() calls can be performed.
+        //
+        // When count_ngram() is called manually, it isn't guaranteed
+        // that all nodes are created that are required for a valid model.
+        // Unigrams in particular may be missing. This can lead to crashes
+        // in filter_candidates or probability calculations.
+        //
+        // For filling models it's best to avoid count_ngram() and
+        // use the safer learn_tokens() instead.
+        virtual bool is_model_valid()
+        {
+            // including removed unigrams with count==0
+            int num_unigrams = ngrams.get_num_children(&ngrams, 0);
+            return num_unigrams == dictionary.get_num_word_types();
+        }
+
         virtual BaseNode* count_ngram(const wchar_t* const* ngram, int n,
                                 int increment=1, bool allow_new_words=true);
         virtual BaseNode* count_ngram(const WordId* wids, int n, int increment);
@@ -803,9 +949,10 @@ class _DynamicModel : public DynamicModelBase
             return ngrams.increment_node_count(node, wids, n, increment);
         }
 
+        // Number of n-grams per level, excluding removed ones with count==0.
         virtual int get_num_ngrams(int level)
-        { 
-            return ngrams.get_num_ngrams(level); 
+        {
+            return ngrams.get_num_ngrams(level);
         }
 
         virtual void reserve_unigrams(int count)
@@ -823,10 +970,19 @@ class _DynamicModel : public DynamicModelBase
         }
 
     protected:
+        // n-gram trie
         TNGRAMS ngrams;
+
+        // smoothing
         Smoothing smoothing;
+
+        // total number of n-grams with exactly one count, per level
         std::vector<int> n1s;
+
+        // total number of n-grams with exactly two counts, per level
         std::vector<int> n2s;
+
+        // discounting parameters for abs. discounting, kneser-ney, per level
         std::vector<double> Ds;
 };
 
